import importlib
import os
import shutil
import subprocess

from git import Repo

from autopack.api import PackResponse, get_pack_details
from autopack.errors import AutoPackError, AutoPackInstallationError
from autopack.get_pack import try_get_pack
from autopack.pack import Pack
from autopack.utils import (find_or_create_autopack_dir, load_metadata_file,
                            write_metadata_file)


def is_dependency_installed(dependency: str) -> bool:
    try:
        importlib.import_module(dependency)
        return True
    except ImportError:
        return False


def install_dependency(dependency: str):
    try:
        subprocess.check_output(
            ["pip", "install", dependency],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        print(f"{dependency} has been successfully installed.")
    except subprocess.CalledProcessError as e:
        print(f"Installation of {dependency} failed with the following error:")
        print(e.output)


def ask_to_install_dependencies(dependencies: list[str], force=False):
    for dependency in dependencies:
        if is_dependency_installed(dependency):
            continue

        if force:
            install_dependency(dependency)
        else:
            print(
                f"This pack requires the dependency {dependency} to be installed. Continue?"
            )
            agree = input("[Yn]")
            if agree.lower() == "y" or agree == "":
                install_dependency(dependency)
            else:
                print(f"Skipping install of {dependency}")


def install_from_git(pack_data: PackResponse) -> str:
    autopack_dir = find_or_create_autopack_dir()

    url = pack_data.repo_url()
    pack_path = os.path.join(autopack_dir, pack_data.pack_path())

    if os.path.exists(pack_path):
        print("Repo already exists, pulling updates")
        Repo(pack_path).remotes.origin.pull()
    else:
        print(f"Cloning repo into {pack_path}")
        Repo.clone_from(url, pack_path)

    return pack_path


def update_metadata_file(pack: PackResponse):
    metadata = load_metadata_file()
    metadata[pack.pack_id] = {
        "pack_id": pack.pack_id,
        "author": pack.author,
        "repo": pack.repo,
        "module_path": pack.module_path,
        "description": pack.description,
        "name": pack.name,
        "dependencies": pack.dependencies,
        "source": pack.source,
        "run_args": pack.run_args,
        "init_args": pack.init_args,
    }

    write_metadata_file(metadata)


def install_pack(pack_id: str, force_dependencies=False) -> Pack:
    print(f"Installing pack: {pack_id}")
    find_or_create_autopack_dir()

    pack = try_get_pack(pack_id, quiet=True)
    if pack:
        print(f"Pack {pack_id} already installed.")
        return pack

    try:
        pack_data = get_pack_details(pack_id, remote=True)

        if not pack_data:
            raise AutoPackInstallationError(f"Could not find pack details")

        ask_to_install_dependencies(pack_data.dependencies, force_dependencies)
    except AutoPackError as e:
        # Maybe do something else
        raise AutoPackInstallationError(f"Could not install pack {e}")
    except BaseException as e:
        raise AutoPackInstallationError(f"Could not install pack {e}")

    git_dir = ""
    try:
        if pack_data.source == "git":
            git_dir = install_from_git(pack_data)

        update_metadata_file(pack_data)
        pack = try_get_pack(pack_id, quiet=True)

        if pack:
            return pack

    except Exception as e:
        raise AutoPackInstallationError(f"Couldn't install pack due to error {e}")

    if git_dir and os.path.isdir(git_dir):
        shutil.rmtree(git_dir)

    raise AutoPackInstallationError(
        "Error: Installation completed but pack could still not be found."
    )



from dataclasses import dataclass
from typing import Any

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class PackResponse:
    """Class to map Pack properties from the API. Internal use only."""

    pack_id: str
    author: str
    repo: str
    module_path: str
    description: str
    name: str
    dependencies: list[str]
    source: str
    run_args: dict[str, Any]
    init_args: dict[str, Any]

    def pack_path(self) -> str:
        return f"{self.author}_{self.repo}_{self.name}".replace("-", "_").replace(
            " ", "_"
        )

    def repo_url(self) -> str:
        return f"https://github.com/{self.author}/{self.repo}.git"



import json

from langchain.chat_models.base import BaseChatModel
from langchain.schema import HumanMessage

from autopack.api import pack_search

TOOL_SELECTION_PROMPT = """
You are an autonomous AI Agent. I have a list of tools to choose from and a task I need you to accomplish. Recommend
tools that would be required for you to complete the task. Return only a comma-separated list of tool_ids and no
other content.

---- TASK ----
{user_input}
---- TOOLS ----
{tools_string}
"""


def select_packs(task_description: str, llm: BaseChatModel) -> list[str]:
    """Given a user input describing the task they wish to accomplish, return a list of Pack IDs that the given LLM
    thinks will be suitable for this task.

    This is good for a "pre-processing" step after receiving the task but before trying to solve it. This allows you
    to benefit from the wide tool selection while keeping your token usage low.

    You can then further filter, install the packs if desired, and then fetch them using get_pack().

    Args:
        task_description (str): A description of the task to be used when selecting tools
        llm (BaseChatModel): An LLM which will be used to evaluate the selection

    Returns:
        list[str]: A list of selected Pack IDs
    """

    packs = pack_search("")
    pack_summaries = []
    for pack in packs:
        pack_summaries.append(
            {
                "tool_id": pack.pack_id,
                "name": pack.name,
                "description": pack.description,
                "arguments": pack.run_args,
            }
        )

    tools_string = json.dumps(pack_summaries)
    prompt = TOOL_SELECTION_PROMPT.format(
        user_input=task_description, tools_string=tools_string
    )

    selected_packs = ask_llm(prompt, llm)

    return selected_packs


def ask_llm(prompt: str, llm: BaseChatModel) -> list[str]:
    """Encapsulate the OpenAI specific stuff to easier support other frameworks in the future"""
    message = HumanMessage(content=prompt)

    response = llm(messages=[message])

    return response.content.split(",")
